"""discover_utils.data

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The discover-utils data package defines data types for streams and annotations as well as providing respective data handlers to load and save data from and to various backends.
"""