"""
Discover-Utils Scripts

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The Discover-Utils Scripts package is a collection of standalone scripts to perform specific machine learning tasks.
It includes a set of Python scripts that allow users to streamline the process of training machine learning models using Discover server modules.

Included Scripts:
    1. train.py - Training Script
    2. extract.py - Extract Script
    3. predict.py - Predict Script
"""