"""discover_utils.data.provider

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The discover-utils data provider package implements different data providers to interact with multiple datastreams and annotations in a coherent way.
"""