"""discover_utils

Authors:
    Dominik Schiller <dominik.schiller@uni-a.de>
    Tobias Hallmen <tobias.hallmen@uni-a.de>
Date:
    06th September 2023

"""

_MAJOR_VERSION = '1'
_MINOR_VERSION = '0'
_PATCH_VERSION = '13'

__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])
